# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Influenciae
-------

The goal of Influenciae is to provide a simple interface to the various influence functions
techniques
"""

__version__ = '0.3.0'

from . import influence
from . import common
from . import rps
from . import trac_in
from . import benchmark
from . import plots
